library angular2.src.change_detection.pipe_transform;

import "package:angular2/src/facade/lang.dart" show BaseException, Type;

/**
 * An interface which all pipes must implement.
 *
 * #Example
 *
 * ```
 * class DoublePipe implements PipeTransform {
 *  transform(value, args = []) {
 *    return `${value}${value}`;
 *  }
 * }
 * ```
 */
abstract class PipeTransform {
  dynamic transform(dynamic value, List<dynamic> args);
}

/**
 * An interface that stateful pipes should implement.
 *
 * #Example
 *
 * ```
 * class StatefulPipe implements PipeTransform, PipeOnDestroy {
 *  connection;
 *
 *  onDestroy() {
 *    this.connection.release();
 *  }
 *
 *  transform(value, args = []) {
 *    this.connection = createConnection();
 *    // ...
 *    return someValue;
 *  }
 * }
 * ```
 */
abstract class PipeOnDestroy {
  void onDestroy();
}
