library angular2.src.core.compiler.element_injector;

import "package:angular2/src/facade/lang.dart"
    show isPresent, isBlank, Type, BaseException, stringify, StringWrapper;
import "package:angular2/src/facade/async.dart"
    show EventEmitter, ObservableWrapper;
import "package:angular2/src/facade/collection.dart"
    show List, ListWrapper, MapWrapper, StringMapWrapper;
import "package:angular2/di.dart"
    show
        Injector,
        ProtoInjector,
        Visibility,
        UNDEFINED,
        Key,
        Dependency,
        bind,
        Binding,
        ResolvedBinding,
        NoBindingError,
        AbstractBindingError,
        CyclicDependencyError,
        resolveForwardRef,
        DependencyProvider;
import "package:angular2/src/di/injector.dart"
    show InjectorInlineStrategy, InjectorDynamicStrategy, BindingWithVisibility;
import "../metadata/di.dart" show AttributeMetadata, QueryMetadata;
import "view.dart" as viewModule;
import "view_manager.dart" as avmModule;
import "view_container_ref.dart" show ViewContainerRef;
import "element_ref.dart" show ElementRef;
import "template_ref.dart" show TemplateRef;
import "../metadata/directives.dart"
    show DirectiveMetadata, ComponentMetadata, LifecycleEvent;
import "directive_lifecycle_reflector.dart" show hasLifecycleHook;
import "package:angular2/src/change_detection/change_detection.dart"
    show ChangeDetector, ChangeDetectorRef;
import "query_list.dart" show QueryList;
import "package:angular2/src/reflection/reflection.dart" show reflector;
import "package:angular2/src/render/api.dart" show RenderDirectiveMetadata;
import "../pipes/pipe_binding.dart" show PipeBinding;

var _staticKeys;

class StaticKeys {
  num viewManagerId;
  num templateRefId;
  num viewContainerId;
  num changeDetectorRefId;
  num elementRefId;
  StaticKeys() {
    this.viewManagerId = Key.get(avmModule.AppViewManager).id;
    this.templateRefId = Key.get(TemplateRef).id;
    this.viewContainerId = Key.get(ViewContainerRef).id;
    this.changeDetectorRefId = Key.get(ChangeDetectorRef).id;
    this.elementRefId = Key.get(ElementRef).id;
  }
  static StaticKeys instance() {
    if (isBlank(_staticKeys)) _staticKeys = new StaticKeys();
    return _staticKeys;
  }
}

class TreeNode<T extends TreeNode<dynamic>> {
  T _parent;
  T _head = null;
  T _tail = null;
  T _next = null;
  TreeNode(T parent) {
    if (isPresent(parent)) parent.addChild(this);
  }
  /**
   * Adds a child to the parent node. The child MUST NOT be a part of a tree.
   */
  void addChild(T child) {
    if (isPresent(this._tail)) {
      this._tail._next = child;
      this._tail = child;
    } else {
      this._tail = this._head = child;
    }
    child._next = null;
    child._parent = this;
  }

  /**
   * Adds a child to the parent node after a given sibling.
   * The child MUST NOT be a part of a tree and the sibling must be present.
   */
  void addChildAfter(T child, T prevSibling) {
    if (isBlank(prevSibling)) {
      var prevHead = this._head;
      this._head = child;
      child._next = prevHead;
      if (isBlank(this._tail)) this._tail = child;
    } else if (isBlank(prevSibling._next)) {
      this.addChild(child);
      return;
    } else {
      child._next = prevSibling._next;
      prevSibling._next = child;
    }
    child._parent = this;
  }

  /**
   * Detaches a node from the parent's tree.
   */
  void remove() {
    if (isBlank(this.parent)) return;
    var nextSibling = this._next;
    var prevSibling = this._findPrev();
    if (isBlank(prevSibling)) {
      this.parent._head = this._next;
    } else {
      prevSibling._next = this._next;
    }
    if (isBlank(nextSibling)) {
      this._parent._tail = prevSibling;
    }
    this._parent = null;
    this._next = null;
  }

  /**
   * Finds a previous sibling or returns null if first child.
   * Assumes the node has a parent.
   * TODO(rado): replace with DoublyLinkedList to avoid O(n) here.
   */
  _findPrev() {
    var node = this.parent._head;
    if (node == this) return null;
    while (!identical(node._next, this)) node = node._next;
    return node;
  }

  get parent {
    return this._parent;
  }

  // TODO(rado): replace with a function call, does too much work for a getter.
  List<T> get children {
    var res = [];
    var child = this._head;
    while (child != null) {
      res.add(child);
      child = child._next;
    }
    return res;
  }
}

class DirectiveDependency extends Dependency {
  String attributeName;
  QueryMetadata queryDecorator;
  DirectiveDependency(
      Key key,
      bool optional,
      Object lowerBoundVisibility,
      Object upperBoundVisibility,
      List<dynamic> properties,
      this.attributeName,
      this.queryDecorator)
      : super(key, optional, lowerBoundVisibility, upperBoundVisibility,
            properties) {
    /* super call moved to initializer */;
    this._verify();
  }
  void _verify() {
    var count = 0;
    if (isPresent(this.queryDecorator)) count++;
    if (isPresent(this.attributeName)) count++;
    if (count > 1) throw new BaseException(
        "A directive injectable can contain only one of the following @Attribute or @Query.");
  }

  static Dependency createFrom(Dependency d) {
    return new DirectiveDependency(
        d.key,
        d.optional,
        d.lowerBoundVisibility,
        d.upperBoundVisibility,
        d.properties,
        DirectiveDependency._attributeName(d.properties),
        DirectiveDependency._query(d.properties));
  }

  static String _attributeName(properties) {
    var p = (ListWrapper.find(properties, (p) => p is AttributeMetadata)
        as AttributeMetadata);
    return isPresent(p) ? p.attributeName : null;
  }

  static QueryMetadata _query(properties) {
    return (ListWrapper.find(properties, (p) => p is QueryMetadata)
        as QueryMetadata);
  }
}

class DirectiveBinding extends ResolvedBinding {
  List<ResolvedBinding> resolvedBindings;
  List<ResolvedBinding> resolvedViewBindings;
  RenderDirectiveMetadata metadata;
  DirectiveBinding(Key key, Function factory, List<Dependency> dependencies,
      this.resolvedBindings, this.resolvedViewBindings, this.metadata)
      : super(key, factory, dependencies) {
    /* super call moved to initializer */;
  }
  bool get callOnDestroy {
    return this.metadata.callOnDestroy;
  }

  bool get callOnChange {
    return this.metadata.callOnChange;
  }

  bool get callOnAllChangesDone {
    return this.metadata.callOnAllChangesDone;
  }

  String get displayName {
    return this.key.displayName;
  }

  List<String> get eventEmitters {
    return isPresent(this.metadata) && isPresent(this.metadata.events)
        ? this.metadata.events
        : [];
  }

  Map<String, String> get hostActions {
    return isPresent(this.metadata) && isPresent(this.metadata.hostActions)
        ? this.metadata.hostActions
        : new Map();
  }

  get changeDetection {
    return this.metadata.changeDetection;
  }

  static DirectiveBinding createFromBinding(
      Binding binding, DirectiveMetadata ann) {
    if (isBlank(ann)) {
      ann = new DirectiveMetadata();
    }
    var rb = binding.resolve();
    var deps = ListWrapper.map(rb.dependencies, DirectiveDependency.createFrom);
    var resolvedBindings =
        isPresent(ann.bindings) ? Injector.resolve(ann.bindings) : [];
    var resolvedViewBindings = ann is ComponentMetadata &&
        isPresent(ann.viewBindings) ? Injector.resolve(ann.viewBindings) : [];
    var metadata = RenderDirectiveMetadata.create(
        id: stringify(rb.key.token),
        type: ann is ComponentMetadata
            ? RenderDirectiveMetadata.COMPONENT_TYPE
            : RenderDirectiveMetadata.DIRECTIVE_TYPE,
        selector: ann.selector,
        compileChildren: ann.compileChildren,
        events: ann.events,
        host: isPresent(ann.host)
            ? MapWrapper.createFromStringMap(ann.host)
            : null,
        properties: ann.properties,
        readAttributes: DirectiveBinding._readAttributes(deps),
        callOnDestroy:
            hasLifecycleHook(LifecycleEvent.onDestroy, rb.key.token, ann),
        callOnChange:
            hasLifecycleHook(LifecycleEvent.onChange, rb.key.token, ann),
        callOnCheck:
            hasLifecycleHook(LifecycleEvent.onCheck, rb.key.token, ann),
        callOnInit: hasLifecycleHook(LifecycleEvent.onInit, rb.key.token, ann),
        callOnAllChangesDone: hasLifecycleHook(
            LifecycleEvent.onAllChangesDone, rb.key.token, ann),
        changeDetection: ann is ComponentMetadata ? ann.changeDetection : null,
        exportAs: ann.exportAs);
    return new DirectiveBinding(rb.key, rb.factory, deps, resolvedBindings,
        resolvedViewBindings, metadata);
  }

  static _readAttributes(deps) {
    var readAttributes = [];
    ListWrapper.forEach(deps, (dep) {
      if (isPresent(dep.attributeName)) {
        readAttributes.add(dep.attributeName);
      }
    });
    return readAttributes;
  }

  static DirectiveBinding createFromType(
      Type type, DirectiveMetadata annotation) {
    var binding = new Binding(type, toClass: type);
    return DirectiveBinding.createFromBinding(binding, annotation);
  }
}

// TODO(rado): benchmark and consider rolling in as ElementInjector fields.
class PreBuiltObjects {
  avmModule.AppViewManager viewManager;
  viewModule.AppView view;
  ElementRef elementRef;
  TemplateRef templateRef;
  PreBuiltObjects(
      this.viewManager, this.view, this.elementRef, this.templateRef) {}
}

class EventEmitterAccessor {
  String eventName;
  Function getter;
  EventEmitterAccessor(this.eventName, this.getter) {}
  Object subscribe(
      viewModule.AppView view, num boundElementIndex, Object directive) {
    var eventEmitter = this.getter(directive);
    return ObservableWrapper.subscribe(
        eventEmitter,
        (eventObj) => view.triggerEventHandlers(
            this.eventName, eventObj, boundElementIndex));
  }
}

class HostActionAccessor {
  String methodName;
  Function getter;
  HostActionAccessor(this.methodName, this.getter) {}
  Object subscribe(
      viewModule.AppView view, num boundElementIndex, Object directive) {
    var eventEmitter = this.getter(directive);
    return ObservableWrapper.subscribe(
        eventEmitter,
        (actionArgs) => view.invokeElementMethod(
            boundElementIndex, this.methodName, actionArgs));
  }
}

List<EventEmitterAccessor> _createEventEmitterAccessors(
    BindingWithVisibility bwv) {
  var binding = bwv.binding;
  if (!(binding is DirectiveBinding)) return [];
  var db = (binding as DirectiveBinding);
  return ListWrapper.map(db.eventEmitters, (eventConfig) {
    var fieldName;
    var eventName;
    var colonIdx = eventConfig.indexOf(":");
    if (colonIdx > -1) {
      // long format: 'fieldName: eventName'
      fieldName = StringWrapper.substring(eventConfig, 0, colonIdx).trim();
      eventName = StringWrapper.substring(eventConfig, colonIdx + 1).trim();
    } else {
      // short format: 'name' when fieldName and eventName are the same
      fieldName = eventName = eventConfig;
    }
    return new EventEmitterAccessor(eventName, reflector.getter(fieldName));
  });
}

List<HostActionAccessor> _createHostActionAccessors(BindingWithVisibility bwv) {
  var binding = bwv.binding;
  if (!(binding is DirectiveBinding)) return [];
  var res = [];
  var db = (binding as DirectiveBinding);
  MapWrapper.forEach(db.hostActions, (actionExpression, actionName) {
    res.add(
        new HostActionAccessor(actionExpression, reflector.getter(actionName)));
  });
  return res;
}

class ProtoElementInjector {
  ProtoElementInjector parent;
  int index;
  num distanceToParent;
  bool _firstBindingIsComponent;
  Map<String, num> directiveVariableBindings;
  viewModule.AppView view;
  Map<String, String> attributes;
  List<List<EventEmitterAccessor>> eventEmitterAccessors;
  List<List<HostActionAccessor>> hostActionAccessors;
  ProtoInjector protoInjector;
  static ProtoElementInjector create(
      ProtoElementInjector parent,
      num index,
      List<ResolvedBinding> bindings,
      bool firstBindingIsComponent,
      num distanceToParent,
      Map<String, num> directiveVariableBindings) {
    var bd = [];
    ProtoElementInjector._createDirectiveBindingWithVisibility(
        bindings, bd, firstBindingIsComponent);
    if (firstBindingIsComponent) {
      ProtoElementInjector._createViewBindingsWithVisibility(bindings, bd);
    }
    ProtoElementInjector._createBindingsWithVisibility(
        bindings, bd, firstBindingIsComponent);
    return new ProtoElementInjector(parent, index, bd, distanceToParent,
        firstBindingIsComponent, directiveVariableBindings);
  }

  static _createDirectiveBindingWithVisibility(
      List<ResolvedBinding> dirBindings,
      List<BindingWithVisibility> bd,
      bool firstBindingIsComponent) {
    ListWrapper.forEach(dirBindings, (dirBinding) {
      bd.add(ProtoElementInjector._createBindingWithVisibility(
          firstBindingIsComponent, dirBinding, dirBindings, dirBinding));
    });
  }

  static _createBindingsWithVisibility(List<ResolvedBinding> dirBindings,
      List<BindingWithVisibility> bd, bool firstBindingIsComponent) {
    ListWrapper.forEach(dirBindings, (dirBinding) {
      ListWrapper.forEach(dirBinding.resolvedBindings, (b) {
        bd.add(ProtoElementInjector._createBindingWithVisibility(
            firstBindingIsComponent, dirBinding, dirBindings, b));
      });
    });
  }

  static _createBindingWithVisibility(
      firstBindingIsComponent, dirBinding, dirBindings, binding) {
    var isComponent =
        firstBindingIsComponent && identical(dirBindings[0], dirBinding);
    return new BindingWithVisibility(
        binding, isComponent ? Visibility.PublicAndPrivate : Visibility.Public);
  }

  static _createViewBindingsWithVisibility(
      List<ResolvedBinding> bindings, List<BindingWithVisibility> bd) {
    var db = (bindings[0] as DirectiveBinding);
    ListWrapper.forEach(db.resolvedViewBindings,
        (b) => bd.add(new BindingWithVisibility(b, Visibility.Private)));
  }

  ProtoElementInjector(
      this.parent,
      this.index,
      List<BindingWithVisibility> bwv,
      this.distanceToParent,
      this._firstBindingIsComponent,
      this.directiveVariableBindings) {
    var length = bwv.length;
    this.protoInjector = new ProtoInjector(bwv);
    this.eventEmitterAccessors = ListWrapper.createFixedSize(length);
    this.hostActionAccessors = ListWrapper.createFixedSize(length);
    for (var i = 0; i < length; ++i) {
      this.eventEmitterAccessors[i] = _createEventEmitterAccessors(bwv[i]);
      this.hostActionAccessors[i] = _createHostActionAccessors(bwv[i]);
    }
  }
  ElementInjector instantiate(ElementInjector parent) {
    return new ElementInjector(this, parent);
  }

  ProtoElementInjector directParent() {
    return this.distanceToParent < 2 ? this.parent : null;
  }

  bool get hasBindings {
    return this.eventEmitterAccessors.length > 0;
  }

  dynamic getBindingAtIndex(num index) {
    return this.protoInjector.getBindingAtIndex(index);
  }
}

class _Context {
  dynamic element;
  dynamic componentElement;
  dynamic injector;
  _Context(this.element, this.componentElement, this.injector) {}
}

class ElementInjector extends TreeNode<ElementInjector>
    implements DependencyProvider {
  ProtoElementInjector _proto;
  ElementInjector _host;
  PreBuiltObjects _preBuiltObjects = null;
  // Queries are added during construction or linking with a new parent.

  // They are removed only through unlinking.
  QueryRef _query0;
  QueryRef _query1;
  QueryRef _query2;
  bool hydrated;
  Injector _injector;
  _ElementInjectorStrategy _strategy;
  ElementInjector(this._proto, ElementInjector parent) : super(parent) {
    /* super call moved to initializer */;
    this._injector = new Injector(
        this._proto.protoInjector, null, this, () => this._debugContext());
    // we couple ourselves to the injector strategy to avoid polymoprhic calls
    var injectorStrategy = (this._injector.internalStrategy as dynamic);
    this._strategy = injectorStrategy is InjectorInlineStrategy
        ? new ElementInjectorInlineStrategy(injectorStrategy, this)
        : new ElementInjectorDynamicStrategy(injectorStrategy, this);
    this.hydrated = false;
    this._buildQueries();
    this._addParentQueries();
  }
  void dehydrate() {
    this.hydrated = false;
    this._host = null;
    this._preBuiltObjects = null;
    this._strategy.callOnDestroy();
    this._strategy.dehydrate();
  }

  void onAllChangesDone() {
    if (isPresent(this._query0) && identical(this._query0.originator, this)) {
      this._query0.list.fireCallbacks();
    }
    if (isPresent(this._query1) && identical(this._query1.originator, this)) {
      this._query1.list.fireCallbacks();
    }
    if (isPresent(this._query2) && identical(this._query2.originator, this)) {
      this._query2.list.fireCallbacks();
    }
  }

  void hydrate(Injector imperativelyCreatedInjector, ElementInjector host,
      PreBuiltObjects preBuiltObjects) {
    this._host = host;
    this._preBuiltObjects = preBuiltObjects;
    if (isPresent(host)) {
      this._addViewQueries(host);
    }
    this._reattachInjectors(imperativelyCreatedInjector);
    this._strategy.hydrate();
    this._addDirectivesToQueries();
    this._addVarBindingsToQueries();
    this.hydrated = true;
    // TODO(rado): optimize this call, if view queries are not moved around,

    // simply appending to the query list is faster than updating.
    this._updateViewQueries();
  }

  _updateViewQueries() {
    if (isPresent(this._query0) && this._query0.isViewQuery) {
      this._query0.update();
    }
    if (isPresent(this._query1) && this._query1.isViewQuery) {
      this._query1.update();
    }
    if (isPresent(this._query2) && this._query2.isViewQuery) {
      this._query2.update();
    }
  }

  dynamic _debugContext() {
    var p = this._preBuiltObjects;
    var index = p.elementRef.boundElementIndex - p.view.elementOffset;
    var c = this._preBuiltObjects.view.getDebugContext(index, null);
    return isPresent(c)
        ? new _Context(c.element, c.componentElement, c.injector)
        : null;
  }

  void _reattachInjectors(Injector imperativelyCreatedInjector) {
    // Dynamically-loaded component in the template. Not a root ElementInjector.
    if (isPresent(this._parent)) {
      if (isPresent(imperativelyCreatedInjector)) {
        // The imperative injector is similar to having an element between

        // the dynamic-loaded component and its parent => no boundaries.
        this._reattachInjector(
            this._injector, imperativelyCreatedInjector, false);
        this._reattachInjector(
            imperativelyCreatedInjector, this._parent._injector, false);
      } else {
        this._reattachInjector(this._injector, this._parent._injector, false);
      }
    } else if (isPresent(this._host)) {
      // The imperative injector is similar to having an element between

      // the dynamic-loaded component and its parent => no boundary between

      // the component and imperativelyCreatedInjector.

      // But since it is a root ElementInjector, we need to create a boundary

      // between imperativelyCreatedInjector and _host.
      if (isPresent(imperativelyCreatedInjector)) {
        this._reattachInjector(
            this._injector, imperativelyCreatedInjector, false);
        this._reattachInjector(
            imperativelyCreatedInjector, this._host._injector, true);
      } else {
        this._reattachInjector(this._injector, this._host._injector, true);
      }
    } else {
      if (isPresent(imperativelyCreatedInjector)) {
        this._reattachInjector(
            this._injector, imperativelyCreatedInjector, true);
      }
    }
  }

  _reattachInjector(
      Injector injector, Injector parentInjector, bool isBoundary) {
    injector.internalStrategy.attach(parentInjector, isBoundary);
  }

  bool hasVariableBinding(String name) {
    var vb = this._proto.directiveVariableBindings;
    return isPresent(vb) && vb.containsKey(name);
  }

  dynamic getVariableBinding(String name) {
    var index = this._proto.directiveVariableBindings[name];
    return isPresent(index)
        ? this.getDirectiveAtIndex((index as num))
        : this.getElementRef();
  }

  dynamic get(dynamic token) {
    return this._injector.get(token);
  }

  bool hasDirective(Type type) {
    return isPresent(this._injector.getOptional(type));
  }

  List<List<EventEmitterAccessor>> getEventEmitterAccessors() {
    return this._proto.eventEmitterAccessors;
  }

  List<List<HostActionAccessor>> getHostActionAccessors() {
    return this._proto.hostActionAccessors;
  }

  Map<String, num> getDirectiveVariableBindings() {
    return this._proto.directiveVariableBindings;
  }

  dynamic getComponent() {
    return this._strategy.getComponent();
  }

  Injector getInjector() {
    return this._injector;
  }

  ElementRef getElementRef() {
    return this._preBuiltObjects.elementRef;
  }

  ViewContainerRef getViewContainerRef() {
    return new ViewContainerRef(
        this._preBuiltObjects.viewManager, this.getElementRef());
  }

  ElementInjector directParent() {
    return this._proto.distanceToParent < 2 ? this.parent : null;
  }

  bool isComponentKey(Key key) {
    return this._strategy.isComponentKey(key);
  }

  dynamic getDependency(
      Injector injector, ResolvedBinding binding, Dependency dep) {
    Key key = dep.key;
    if (binding is DirectiveBinding) {
      var dirDep = (dep as DirectiveDependency);
      var dirBin = binding;
      var staticKeys = StaticKeys.instance();
      if (identical(key.id, staticKeys.viewManagerId)) return this
          ._preBuiltObjects
          .viewManager;
      if (isPresent(dirDep.attributeName)) return this._buildAttribute(dirDep);
      if (isPresent(dirDep.queryDecorator)) return this
          ._findQuery(dirDep.queryDecorator)
          .list;
      if (identical(dirDep.key.id, StaticKeys.instance().changeDetectorRefId)) {
        // We provide the component's view change detector to components and

        // the surrounding component's change detector to directives.
        if (identical(
            dirBin.metadata.type, RenderDirectiveMetadata.COMPONENT_TYPE)) {
          var componentView = this._preBuiltObjects.view.getNestedView(
              this._preBuiltObjects.elementRef.boundElementIndex);
          return componentView.changeDetector.ref;
        } else {
          return this._preBuiltObjects.view.changeDetector.ref;
        }
      }
      if (identical(dirDep.key.id, StaticKeys.instance().elementRefId)) {
        return this.getElementRef();
      }
      if (identical(dirDep.key.id, StaticKeys.instance().viewContainerId)) {
        return this.getViewContainerRef();
      }
      if (identical(dirDep.key.id, StaticKeys.instance().templateRefId)) {
        if (isBlank(this._preBuiltObjects.templateRef)) {
          if (dirDep.optional) {
            return null;
          }
          throw new NoBindingError(null, dirDep.key);
        }
        return this._preBuiltObjects.templateRef;
      }
    } else if (binding is PipeBinding) {
      if (identical(dep.key.id, StaticKeys.instance().changeDetectorRefId)) {
        var componentView = this
            ._preBuiltObjects
            .view
            .getNestedView(this._preBuiltObjects.elementRef.boundElementIndex);
        return componentView.changeDetector.ref;
      }
    }
    return UNDEFINED;
  }

  String _buildAttribute(DirectiveDependency dep) {
    var attributes = this._proto.attributes;
    if (isPresent(attributes) && attributes.containsKey(dep.attributeName)) {
      return attributes[dep.attributeName];
    } else {
      return null;
    }
  }

  void _buildQueriesForDeps(List<DirectiveDependency> deps) {
    for (var i = 0; i < deps.length; i++) {
      var dep = deps[i];
      if (isPresent(dep.queryDecorator)) {
        this._createQueryRef(dep.queryDecorator);
      }
    }
  }

  void _addViewQueries(ElementInjector host) {
    this._addViewQuery(host._query0, host);
    this._addViewQuery(host._query1, host);
    this._addViewQuery(host._query2, host);
  }

  void _addViewQuery(QueryRef queryRef, ElementInjector host) {
    if (isBlank(queryRef) ||
        !queryRef.isViewQuery ||
        this._hasQuery(queryRef)) return;
    if (host._query0.originator == host) {
      // TODO(rado): Replace this.parent check with distanceToParent = 1 when

      // https://github.com/angular/angular/issues/2707 is fixed.
      if (!queryRef.query.descendants && isPresent(this.parent)) return;
      this._assignQueryRef(queryRef);
    }
  }

  void _addVarBindingsToQueries() {
    this._addVarBindingsToQuery(this._query0);
    this._addVarBindingsToQuery(this._query1);
    this._addVarBindingsToQuery(this._query2);
  }

  void _addDirectivesToQueries() {
    this._addDirectivesToQuery(this._query0);
    this._addDirectivesToQuery(this._query1);
    this._addDirectivesToQuery(this._query2);
  }

  void _addVarBindingsToQuery(QueryRef queryRef) {
    if (isBlank(queryRef) || !queryRef.query.isVarBindingQuery) return;
    var vb = queryRef.query.varBindings;
    for (var i = 0; i < vb.length; ++i) {
      if (this.hasVariableBinding(vb[i])) {
        queryRef.list.add(this.getVariableBinding(vb[i]));
      }
    }
  }

  void _addDirectivesToQuery(QueryRef queryRef) {
    if (isBlank(queryRef) || queryRef.query.isVarBindingQuery) return;
    if (queryRef.isViewQuery && queryRef.originator == this) return;
    var matched = [];
    this.addDirectivesMatchingQuery(queryRef.query, matched);
    matched.forEach((s) => queryRef.list.add(s));
  }

  void _createQueryRef(QueryMetadata query) {
    var queryList = new QueryList<dynamic>();
    if (isBlank(this._query0)) {
      this._query0 = new QueryRef(query, queryList, this);
    } else if (isBlank(this._query1)) {
      this._query1 = new QueryRef(query, queryList, this);
    } else if (isBlank(this._query2)) {
      this._query2 = new QueryRef(query, queryList, this);
    } else {
      throw new QueryError();
    }
  }

  void addDirectivesMatchingQuery(QueryMetadata query, List<dynamic> list) {
    var templateRef = this._preBuiltObjects.templateRef;
    if (identical(query.selector, TemplateRef) && isPresent(templateRef)) {
      list.add(templateRef);
    }
    this._strategy.addDirectivesMatchingQuery(query, list);
  }

  void _buildQueries() {
    if (isPresent(this._proto)) {
      this._strategy.buildQueries();
    }
  }

  QueryRef _findQuery(query) {
    if (isPresent(this._query0) && identical(this._query0.query, query)) {
      return this._query0;
    }
    if (isPresent(this._query1) && identical(this._query1.query, query)) {
      return this._query1;
    }
    if (isPresent(this._query2) && identical(this._query2.query, query)) {
      return this._query2;
    }
    throw new BaseException('''Cannot find query for directive ${ query}.''');
  }

  bool _hasQuery(QueryRef query) {
    return this._query0 == query ||
        this._query1 == query ||
        this._query2 == query;
  }

  void link(ElementInjector parent) {
    parent.addChild(this);
    this._addParentQueries();
  }

  void linkAfter(ElementInjector parent, ElementInjector prevSibling) {
    parent.addChildAfter(this, prevSibling);
    this._addParentQueries();
  }

  void unlink() {
    var parent = this.parent;
    this.remove();
    this._removeParentQueries(parent);
  }

  void _addParentQueries() {
    if (isBlank(this.parent)) return;
    this._addParentQuery(this.parent._query0);
    this._addParentQuery(this.parent._query1);
    this._addParentQuery(this.parent._query2);
  }

  void _addParentQuery(query) {
    if (isPresent(query) && !this._hasQuery(query)) {
      this._addQueryToTree(query);
      if (this.hydrated) query.update();
    }
  }

  void _removeParentQueries(ElementInjector parent) {
    this._removeParentQuery(parent._query0);
    this._removeParentQuery(parent._query1);
    this._removeParentQuery(parent._query2);
  }

  _removeParentQuery(QueryRef query) {
    if (isPresent(query)) {
      this._pruneQueryFromTree(query);
      query.update();
    }
  }

  void _pruneQueryFromTree(QueryRef query) {
    this._removeQueryRef(query);
    var child = this._head;
    while (isPresent(child)) {
      child._pruneQueryFromTree(query);
      child = child._next;
    }
  }

  void _addQueryToTree(QueryRef queryRef) {
    if (queryRef.query.descendants == false) {
      if (this == queryRef.originator) {
        this._addQueryToTreeSelfAndRecurse(queryRef);
      } else if (this.parent == queryRef.originator) {
        this._assignQueryRef(queryRef);
      }
    } else {
      this._addQueryToTreeSelfAndRecurse(queryRef);
    }
  }

  void _addQueryToTreeSelfAndRecurse(QueryRef queryRef) {
    this._assignQueryRef(queryRef);
    var child = this._head;
    while (isPresent(child)) {
      child._addQueryToTree(queryRef);
      child = child._next;
    }
  }

  void _assignQueryRef(QueryRef query) {
    if (isBlank(this._query0)) {
      this._query0 = query;
      return;
    } else if (isBlank(this._query1)) {
      this._query1 = query;
      return;
    } else if (isBlank(this._query2)) {
      this._query2 = query;
      return;
    }
    throw new QueryError();
  }

  void _removeQueryRef(QueryRef query) {
    if (this._query0 == query) this._query0 = null;
    if (this._query1 == query) this._query1 = null;
    if (this._query2 == query) this._query2 = null;
  }

  dynamic getDirectiveAtIndex(num index) {
    return this._injector.getAt(index);
  }

  bool hasInstances() {
    return this._proto.hasBindings && this.hydrated;
  }

  ElementInjector getHost() {
    return this._host;
  }

  num getBoundElementIndex() {
    return this._proto.index;
  }

  List<ElementInjector> getRootViewInjectors() {
    var view = this._preBuiltObjects.view;
    return view
        .getNestedView(view.elementOffset + this.getBoundElementIndex())
        .rootElementInjectors;
  }
}

abstract class _ElementInjectorStrategy {
  void callOnDestroy();
  dynamic getComponent();
  bool isComponentKey(Key key);
  void buildQueries();
  void addDirectivesMatchingQuery(QueryMetadata q, List<dynamic> res);
  DirectiveBinding getComponentBinding();
  void hydrate();
  void dehydrate();
}

/**
 * Strategy used by the `ElementInjector` when the number of bindings is 10 or less.
 * In such a case, inlining fields is benefitial for performances.
 */
class ElementInjectorInlineStrategy implements _ElementInjectorStrategy {
  InjectorInlineStrategy injectorStrategy;
  ElementInjector _ei;
  ElementInjectorInlineStrategy(this.injectorStrategy, this._ei) {}
  void hydrate() {
    var i = this.injectorStrategy;
    var p = i.protoStrategy;
    i.resetConstructionCounter();
    if (p.binding0 is DirectiveBinding &&
        isPresent(p.keyId0) &&
        identical(i.obj0, UNDEFINED)) i.obj0 =
        i.instantiateBinding(p.binding0, p.visibility0);
    if (p.binding1 is DirectiveBinding &&
        isPresent(p.keyId1) &&
        identical(i.obj1, UNDEFINED)) i.obj1 =
        i.instantiateBinding(p.binding1, p.visibility1);
    if (p.binding2 is DirectiveBinding &&
        isPresent(p.keyId2) &&
        identical(i.obj2, UNDEFINED)) i.obj2 =
        i.instantiateBinding(p.binding2, p.visibility2);
    if (p.binding3 is DirectiveBinding &&
        isPresent(p.keyId3) &&
        identical(i.obj3, UNDEFINED)) i.obj3 =
        i.instantiateBinding(p.binding3, p.visibility3);
    if (p.binding4 is DirectiveBinding &&
        isPresent(p.keyId4) &&
        identical(i.obj4, UNDEFINED)) i.obj4 =
        i.instantiateBinding(p.binding4, p.visibility4);
    if (p.binding5 is DirectiveBinding &&
        isPresent(p.keyId5) &&
        identical(i.obj5, UNDEFINED)) i.obj5 =
        i.instantiateBinding(p.binding5, p.visibility5);
    if (p.binding6 is DirectiveBinding &&
        isPresent(p.keyId6) &&
        identical(i.obj6, UNDEFINED)) i.obj6 =
        i.instantiateBinding(p.binding6, p.visibility6);
    if (p.binding7 is DirectiveBinding &&
        isPresent(p.keyId7) &&
        identical(i.obj7, UNDEFINED)) i.obj7 =
        i.instantiateBinding(p.binding7, p.visibility7);
    if (p.binding8 is DirectiveBinding &&
        isPresent(p.keyId8) &&
        identical(i.obj8, UNDEFINED)) i.obj8 =
        i.instantiateBinding(p.binding8, p.visibility8);
    if (p.binding9 is DirectiveBinding &&
        isPresent(p.keyId9) &&
        identical(i.obj9, UNDEFINED)) i.obj9 =
        i.instantiateBinding(p.binding9, p.visibility9);
  }

  dehydrate() {
    var i = this.injectorStrategy;
    i.obj0 = UNDEFINED;
    i.obj1 = UNDEFINED;
    i.obj2 = UNDEFINED;
    i.obj3 = UNDEFINED;
    i.obj4 = UNDEFINED;
    i.obj5 = UNDEFINED;
    i.obj6 = UNDEFINED;
    i.obj7 = UNDEFINED;
    i.obj8 = UNDEFINED;
    i.obj9 = UNDEFINED;
  }

  void callOnDestroy() {
    var i = this.injectorStrategy;
    var p = i.protoStrategy;
    if (p.binding0 is DirectiveBinding &&
        ((p.binding0 as DirectiveBinding)).callOnDestroy) {
      i.obj0.onDestroy();
    }
    if (p.binding1 is DirectiveBinding &&
        ((p.binding1 as DirectiveBinding)).callOnDestroy) {
      i.obj1.onDestroy();
    }
    if (p.binding2 is DirectiveBinding &&
        ((p.binding2 as DirectiveBinding)).callOnDestroy) {
      i.obj2.onDestroy();
    }
    if (p.binding3 is DirectiveBinding &&
        ((p.binding3 as DirectiveBinding)).callOnDestroy) {
      i.obj3.onDestroy();
    }
    if (p.binding4 is DirectiveBinding &&
        ((p.binding4 as DirectiveBinding)).callOnDestroy) {
      i.obj4.onDestroy();
    }
    if (p.binding5 is DirectiveBinding &&
        ((p.binding5 as DirectiveBinding)).callOnDestroy) {
      i.obj5.onDestroy();
    }
    if (p.binding6 is DirectiveBinding &&
        ((p.binding6 as DirectiveBinding)).callOnDestroy) {
      i.obj6.onDestroy();
    }
    if (p.binding7 is DirectiveBinding &&
        ((p.binding7 as DirectiveBinding)).callOnDestroy) {
      i.obj7.onDestroy();
    }
    if (p.binding8 is DirectiveBinding &&
        ((p.binding8 as DirectiveBinding)).callOnDestroy) {
      i.obj8.onDestroy();
    }
    if (p.binding9 is DirectiveBinding &&
        ((p.binding9 as DirectiveBinding)).callOnDestroy) {
      i.obj9.onDestroy();
    }
  }

  dynamic getComponent() {
    return this.injectorStrategy.obj0;
  }

  bool isComponentKey(Key key) {
    return this._ei._proto._firstBindingIsComponent &&
        isPresent(key) &&
        identical(key.id, this.injectorStrategy.protoStrategy.keyId0);
  }

  void buildQueries() {
    var p = this.injectorStrategy.protoStrategy;
    if (p.binding0 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding0.dependencies as List<DirectiveDependency>));
    }
    if (p.binding1 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding1.dependencies as List<DirectiveDependency>));
    }
    if (p.binding2 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding2.dependencies as List<DirectiveDependency>));
    }
    if (p.binding3 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding3.dependencies as List<DirectiveDependency>));
    }
    if (p.binding4 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding4.dependencies as List<DirectiveDependency>));
    }
    if (p.binding5 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding5.dependencies as List<DirectiveDependency>));
    }
    if (p.binding6 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding6.dependencies as List<DirectiveDependency>));
    }
    if (p.binding7 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding7.dependencies as List<DirectiveDependency>));
    }
    if (p.binding8 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding8.dependencies as List<DirectiveDependency>));
    }
    if (p.binding9 is DirectiveBinding) {
      this._ei._buildQueriesForDeps(
          (p.binding9.dependencies as List<DirectiveDependency>));
    }
  }

  void addDirectivesMatchingQuery(QueryMetadata query, List<dynamic> list) {
    var i = this.injectorStrategy;
    var p = i.protoStrategy;
    if (isPresent(p.binding0) &&
        identical(p.binding0.key.token, query.selector)) {
      if (identical(i.obj0, UNDEFINED)) i.obj0 =
          i.instantiateBinding(p.binding0, p.visibility0);
      list.add(i.obj0);
    }
    if (isPresent(p.binding1) &&
        identical(p.binding1.key.token, query.selector)) {
      if (identical(i.obj1, UNDEFINED)) i.obj1 =
          i.instantiateBinding(p.binding1, p.visibility1);
      list.add(i.obj1);
    }
    if (isPresent(p.binding2) &&
        identical(p.binding2.key.token, query.selector)) {
      if (identical(i.obj2, UNDEFINED)) i.obj2 =
          i.instantiateBinding(p.binding2, p.visibility2);
      list.add(i.obj2);
    }
    if (isPresent(p.binding3) &&
        identical(p.binding3.key.token, query.selector)) {
      if (identical(i.obj3, UNDEFINED)) i.obj3 =
          i.instantiateBinding(p.binding3, p.visibility3);
      list.add(i.obj3);
    }
    if (isPresent(p.binding4) &&
        identical(p.binding4.key.token, query.selector)) {
      if (identical(i.obj4, UNDEFINED)) i.obj4 =
          i.instantiateBinding(p.binding4, p.visibility4);
      list.add(i.obj4);
    }
    if (isPresent(p.binding5) &&
        identical(p.binding5.key.token, query.selector)) {
      if (identical(i.obj5, UNDEFINED)) i.obj5 =
          i.instantiateBinding(p.binding5, p.visibility5);
      list.add(i.obj5);
    }
    if (isPresent(p.binding6) &&
        identical(p.binding6.key.token, query.selector)) {
      if (identical(i.obj6, UNDEFINED)) i.obj6 =
          i.instantiateBinding(p.binding6, p.visibility6);
      list.add(i.obj6);
    }
    if (isPresent(p.binding7) &&
        identical(p.binding7.key.token, query.selector)) {
      if (identical(i.obj7, UNDEFINED)) i.obj7 =
          i.instantiateBinding(p.binding7, p.visibility7);
      list.add(i.obj7);
    }
    if (isPresent(p.binding8) &&
        identical(p.binding8.key.token, query.selector)) {
      if (identical(i.obj8, UNDEFINED)) i.obj8 =
          i.instantiateBinding(p.binding8, p.visibility8);
      list.add(i.obj8);
    }
    if (isPresent(p.binding9) &&
        identical(p.binding9.key.token, query.selector)) {
      if (identical(i.obj9, UNDEFINED)) i.obj9 =
          i.instantiateBinding(p.binding9, p.visibility9);
      list.add(i.obj9);
    }
  }

  DirectiveBinding getComponentBinding() {
    var p = this.injectorStrategy.protoStrategy;
    return (p.binding0 as DirectiveBinding);
  }
}

/**
 * Strategy used by the `ElementInjector` when the number of bindings is 10 or less.
 * In such a case, inlining fields is benefitial for performances.
 */
class ElementInjectorDynamicStrategy implements _ElementInjectorStrategy {
  InjectorDynamicStrategy injectorStrategy;
  ElementInjector _ei;
  ElementInjectorDynamicStrategy(this.injectorStrategy, this._ei) {}
  void hydrate() {
    var inj = this.injectorStrategy;
    var p = inj.protoStrategy;
    for (var i = 0; i < p.keyIds.length; i++) {
      if (p.bindings[i] is DirectiveBinding &&
          isPresent(p.keyIds[i]) &&
          identical(inj.objs[i], UNDEFINED)) {
        inj.objs[i] = inj.instantiateBinding(p.bindings[i], p.visibilities[i]);
      }
    }
  }

  void dehydrate() {
    var inj = this.injectorStrategy;
    ListWrapper.fill(inj.objs, UNDEFINED);
  }

  void callOnDestroy() {
    var ist = this.injectorStrategy;
    var p = ist.protoStrategy;
    for (var i = 0; i < p.bindings.length; i++) {
      if (p.bindings[i] is DirectiveBinding &&
          ((p.bindings[i] as DirectiveBinding)).callOnDestroy) {
        ist.objs[i].onDestroy();
      }
    }
  }

  dynamic getComponent() {
    return this.injectorStrategy.objs[0];
  }

  bool isComponentKey(Key key) {
    var p = this.injectorStrategy.protoStrategy;
    return this._ei._proto._firstBindingIsComponent &&
        isPresent(key) &&
        identical(key.id, p.keyIds[0]);
  }

  void buildQueries() {
    var inj = this.injectorStrategy;
    var p = inj.protoStrategy;
    for (var i = 0; i < p.bindings.length; i++) {
      if (p.bindings[i] is DirectiveBinding) {
        this._ei._buildQueriesForDeps(
            (p.bindings[i].dependencies as List<DirectiveDependency>));
      }
    }
  }

  void addDirectivesMatchingQuery(QueryMetadata query, List<dynamic> list) {
    var ist = this.injectorStrategy;
    var p = ist.protoStrategy;
    for (var i = 0; i < p.bindings.length; i++) {
      if (identical(p.bindings[i].key.token, query.selector)) {
        if (identical(ist.objs[i], UNDEFINED)) {
          ist.objs[i] =
              ist.instantiateBinding(p.bindings[i], p.visibilities[i]);
        }
        list.add(ist.objs[i]);
      }
    }
  }

  DirectiveBinding getComponentBinding() {
    var p = this.injectorStrategy.protoStrategy;
    return (p.bindings[0] as DirectiveBinding);
  }
}

class QueryError extends BaseException {
  String message;
  // TODO(rado): pass the names of the active directives.
  QueryError() : super() {
    /* super call moved to initializer */;
    this.message = "Only 3 queries can be concurrently active in a template.";
  }
  String toString() {
    return this.message;
  }
}

class QueryRef {
  QueryMetadata query;
  QueryList<dynamic> list;
  ElementInjector originator;
  QueryRef(this.query, this.list, this.originator) {}
  bool get isViewQuery {
    return this.query.isViewQuery;
  }

  void update() {
    var aggregator = [];
    if (this.query.isViewQuery) {
      // intentionally skipping originator for view queries.
      var rootViewInjectors = this.originator.getRootViewInjectors();
      for (var i = 0; i < rootViewInjectors.length; i++) {
        this.visit(rootViewInjectors[i], aggregator);
      }
    } else {
      this.visit(this.originator, aggregator);
    }
    this.list.reset(aggregator);
  }

  void visit(ElementInjector inj, List<dynamic> aggregator) {
    if (isBlank(inj) || !inj._hasQuery(this) || !inj.hydrated) return;
    if (this.query.isVarBindingQuery) {
      this._aggregateVariableBindings(inj, aggregator);
    } else {
      this._aggregateDirective(inj, aggregator);
    }
    var child = inj._head;
    while (isPresent(child)) {
      this.visit(child, aggregator);
      child = child._next;
    }
  }

  void _aggregateVariableBindings(
      ElementInjector inj, List<dynamic> aggregator) {
    var vb = this.query.varBindings;
    for (var i = 0; i < vb.length; ++i) {
      if (inj.hasVariableBinding(vb[i])) {
        aggregator.add(inj.getVariableBinding(vb[i]));
      }
    }
  }

  void _aggregateDirective(ElementInjector inj, List<dynamic> aggregator) {
    inj.addDirectivesMatchingQuery(this.query, aggregator);
  }
}
