library angular2.src.core.compiler.pipe_resolver;

import "package:angular2/di.dart" show resolveForwardRef, Injectable;
import "package:angular2/src/facade/lang.dart"
    show Type, isPresent, BaseException, stringify;
import "../metadata/directives.dart" show PipeMetadata;
import "package:angular2/src/reflection/reflection.dart" show reflector;

/**
 * Resolve a `Type` for [PipeMetadata].
 *
 * This interface can be overridden by the application developer to create custom behavior.
 *
 * See [Compiler]
 */
@Injectable()
class PipeResolver {
  /**
   * Return [PipeMetadata] for a given `Type`.
   */
  PipeMetadata resolve(Type type) {
    var metas = reflector.annotations(resolveForwardRef(type));
    if (isPresent(metas)) {
      for (var i = 0; i < metas.length; i++) {
        var annotation = metas[i];
        if (annotation is PipeMetadata) {
          return annotation;
        }
      }
    }
    throw new BaseException(
        '''No Pipe decorator found on ${ stringify ( type )}''');
  }
}
