library angular2.src.core.metadata.view;

import "package:angular2/src/facade/lang.dart" show Type;
import "package:angular2/src/render/api.dart" show ViewEncapsulation;
export "package:angular2/src/render/api.dart" show ViewEncapsulation;

/**
 * Declares the available HTML templates for an application.
 *
 * Each angular component requires a single `@Component` and at least one `@View` annotation. The
 * `@View` annotation specifies the HTML template to use, and lists the directives that are active
 * within the template.
 *
 * When a component is instantiated, the template is loaded into the component's shadow root, and
 * the expressions and statements in the template are evaluated against the component.
 *
 * For details on the `@Component` annotation, see [ComponentMetadata].
 *
 * ## Example
 *
 * ```
 * @Component({
 *   selector: 'greet'
 * })
 * @View({
 *   template: 'Hello {{name}}!',
 *   directives: [GreetUser, Bold]
 * })
 * class Greet {
 *   name: string;
 *
 *   constructor() {
 *     this.name = 'World';
 *   }
 * }
 * ```
 */
class ViewMetadata {
  /**
   * Specifies a template URL for an angular component.
   *
   * NOTE: either `templateUrl` or `template` should be used, but not both.
   */
  final String templateUrl;
  /**
   * Specifies an inline template for an angular component.
   *
   * NOTE: either `templateUrl` or `template` should be used, but not both.
   */
  final String template;
  /**
   * Specifies stylesheet URLs for an angular component.
   */
  final List<String> styleUrls;
  /**
   * Specifies an inline stylesheet for an angular component.
   */
  final List<String> styles;
  /**
   * Specifies a list of directives that can be used within a template.
   *
   * Directives must be listed explicitly to provide proper component encapsulation.
   *
   * ## Example
   *
   * ```javascript
   * @Component({
   *     selector: 'my-component'
   *   })
   * @View({
   *   directives: [For]
   *   template: '
   *   <ul>
   *     <li *ng-for="#item of items">{{item}}</li>
   *   </ul>'
   * })
   * class MyComponent {
   * }
   * ```
   */

  // TODO(tbosch): use Type | Binding | List<any> when Dart supports union types,

  // as otherwise we would need to import Binding type and Dart would warn

  // for an unused import.
  final List<dynamic /* Type | dynamic | List < dynamic > */ > directives;
  final List<dynamic /* Type | dynamic | List < dynamic > */ > pipes;
  /**
   * Specify how the template and the styles should be encapsulated.
   * The default is [ViewEncapsulation#EMULATED `ViewEncapsulation.EMULATED`] if the view
   * has styles,
   * otherwise [ViewEncapsulation#NONE `ViewEncapsulation.NONE`].
   */
  final ViewEncapsulation encapsulation;
  const ViewMetadata(
      {templateUrl,
      template,
      directives,
      pipes,
      encapsulation,
      styles,
      styleUrls})
      : templateUrl = templateUrl,
        template = template,
        styleUrls = styleUrls,
        styles = styles,
        directives = directives,
        pipes = pipes,
        encapsulation = encapsulation;
}
