library angular2.src.router.route_config_impl;

import "package:angular2/src/facade/lang.dart" show Type;
import "package:angular2/src/facade/collection.dart" show List;
import "route_definition.dart" show RouteDefinition;
export "route_definition.dart" show RouteDefinition;
import "package:angular2/di.dart" show OpaqueToken;

const OpaqueToken ROUTE_DATA = const OpaqueToken("routeData");

/**
 * You use the RouteConfig annotation to add routes to a component.
 *
 * Supported keys:
 * - `path` (required)
 * - `component`, `loader`,  `redirectTo` (requires exactly one of these)
 * - `as` (optional)
 * - `data` (optional)
 */
class RouteConfig {
  final List<RouteDefinition> configs;
  const RouteConfig(this.configs);
}

class Route implements RouteDefinition {
  final dynamic data;
  final String path;
  final Type component;
  final String as;
  // added next two properties to work around https://github.com/Microsoft/TypeScript/issues/4107
  final Function loader;
  final String redirectTo;
  const Route({path, component, as, data})
      : path = path,
        component = component,
        as = as,
        loader = null,
        redirectTo = null,
        data = data;
}

class AuxRoute implements RouteDefinition {
  final dynamic data = null;
  final String path;
  final Type component;
  final String as;
  // added next two properties to work around https://github.com/Microsoft/TypeScript/issues/4107
  final Function loader = null;
  final String redirectTo = null;
  const AuxRoute({path, component, as})
      : path = path,
        component = component,
        as = as;
}

class AsyncRoute implements RouteDefinition {
  final dynamic data;
  final String path;
  final Function loader;
  final String as;
  const AsyncRoute({path, loader, as, data})
      : path = path,
        loader = loader,
        as = as,
        data = data;
}

class Redirect implements RouteDefinition {
  final String path;
  final String redirectTo;
  final String as = null;
  // added next property to work around https://github.com/Microsoft/TypeScript/issues/4107
  final Function loader = null;
  final dynamic data = null;
  const Redirect({path, redirectTo})
      : path = path,
        redirectTo = redirectTo;
}
