library angular2.test.core.pipes.pipes_spec;

import "package:angular2/test_lib.dart"
    show
        ddescribe,
        xdescribe,
        describe,
        it,
        iit,
        xit,
        expect,
        beforeEach,
        afterEach;
import "package:angular2/change_detection.dart" show PipeTransform;
import "package:angular2/di.dart" show Injector, Inject, bind;
import "package:angular2/src/core/pipes/pipes.dart" show ProtoPipes, Pipes;
import "package:angular2/src/core/pipes/pipe_binding.dart" show PipeBinding;
import "package:angular2/metadata.dart" show Pipe;

class PipeA implements PipeTransform {
  transform(a, b) {}
  onDestroy() {}
}

class PipeB implements PipeTransform {
  var dep;
  PipeB(@Inject("dep") dynamic dep) {
    this.dep = dep;
  }
  transform(a, b) {}
  onDestroy() {}
}

main() {
  describe("Pipes", () {
    var injector;
    beforeEach(() {
      injector = Injector.resolveAndCreate([bind("dep").toValue("dependency")]);
    });
    it("should instantiate a pipe", () {
      var proto = new ProtoPipes(
          [PipeBinding.createFromType(PipeA, new Pipe(name: "a"))]);
      var pipes = new Pipes(proto, injector);
      expect(pipes.get("a")).toBeAnInstanceOf(PipeA);
    });
    it("should throw when no pipe found", () {
      var proto = new ProtoPipes([]);
      var pipes = new Pipes(proto, injector);
      expect(() => pipes.get("invalid"))
          .toThrowErrorWith("Cannot find pipe 'invalid'");
    });
    it("should inject dependencies from the provided injector", () {
      var proto = new ProtoPipes(
          [PipeBinding.createFromType(PipeB, new Pipe(name: "b"))]);
      var pipes = new Pipes(proto, injector);
      expect(pipes.get("b").dep).toEqual("dependency");
    });
  });
}
