library angular2.test.web_workers.worker.xhr_impl_spec;

import "package:angular2/test_lib.dart"
    show
        AsyncTestCompleter,
        inject,
        describe,
        it,
        expect,
        beforeEach,
        createTestInjector,
        beforeEachBindings,
        SpyObject,
        proxy;
import "package:angular2/src/facade/lang.dart" show Type;
import "package:angular2/src/web-workers/worker/broker.dart"
    show MessageBroker, UiArguments;
import "package:angular2/src/web-workers/worker/xhr_impl.dart"
    show WebWorkerXHRImpl;
import "package:angular2/src/facade/async.dart" show PromiseWrapper;

main() {
  describe("WebWorkerXHRImpl", () {
    it(
        "should pass requests through the broker and return the response",
        inject([AsyncTestCompleter], (async) {
          const URL = "http://www.example.com/test";
          const RESPONSE = "Example response text";
          dynamic messageBroker = new SpyMessageBroker();
          messageBroker
              .spy("runOnUiThread")
              .andCallFake((UiArguments args, Type returnType) {
            expect(args.type).toEqual("xhr");
            expect(args.method).toEqual("get");
            expect(args.args.length).toEqual(1);
            expect(args.args[0].value).toEqual(URL);
            return PromiseWrapper.wrap(() {
              return RESPONSE;
            });
          });
          var xhrImpl = new WebWorkerXHRImpl(messageBroker);
          xhrImpl.get(URL).then((response) {
            expect(response).toEqual(RESPONSE);
            async.done();
          });
        }));
  });
}

@proxy
class SpyMessageBroker extends SpyObject implements MessageBroker {
  SpyMessageBroker() : super(MessageBroker) {
    /* super call moved to initializer */;
  }
  noSuchMethod(m) {
    return super.noSuchMethod(m);
  }
}
